#include "datatype.h"
unsigned char far* screen1;
unsigned char far* screen2;
unsigned char far* screen3;
unsigned char far* env_map;


#include "common.h"
#include "initcode.h"

inline void rotate_vertices(vertex*,
		long,double,double,double);
void rotate_Vnorm(vertex*,long);
void draw_faces(sort_face*);
void triangle(vertex*,vertex*,vertex*);

double Hx,Hy,Hz;


int main(void)
{
	/* Let's allocate some memory for our structures and data */
	screen1 = (unsigned char far*) farcalloc(64000l,1);
	screen2 = (unsigned char far*) farcalloc(64500l,1);
	screen3 = (unsigned char far*) farcalloc(64500l,1);
	env_map = (unsigned char far*) farcalloc(65535l,1);
	if(!screen1) error("not enought memory");
	if(!screen2) error("not enought memory");
	if(!screen3) error("not enought memory");
	if(!env_map) error("not enought memory");

	const long max_vertices = 65000l/sizeof(vertex);
		// maximum amount of vertices that can be stored in memory
	const long max_faces = 65000l/sizeof(face);
		// maximum amount of faces that can be stored in memory


	vertex far* vertices = (vertex far*)( farcalloc(max_vertices,sizeof(vertex) ) );
		/* mem for all the vertices */

	face far* faces = (face far*)
					( farcalloc(max_faces,sizeof(face) ) );
		/* and faces */

	sort_face far* faces_for_drawing = (sort_face far*)
					( farcalloc(max_faces+1,sizeof(sort_face) ) );
		/* this is necessary when sorting faces */
	if(!vertices)  error("not enought memory");
	if(!faces)  error("not enought memory");
	if(!faces_for_drawing)  error("not enought memory");

	long no_of_vertices=0,no_of_faces=0;

   read_config();
	if(objraw)
		read_raw(vertices,max_vertices,faces,max_faces,
					no_of_vertices,no_of_faces);
	/* function modifies the variables no_of_faces and no_of_vertices
		while reading data from file */
	else
		read_file(vertices,max_vertices,faces,max_faces,
					no_of_vertices,no_of_faces);

	read_picture(screen3);
		/* reads picture for background to screen3 pointer */

	if ( !no_of_vertices || !no_of_faces ) error("plik uszkodzony");
		// if no vertices of no faces read
	init_gfx();

	setup_palette();	// must be after initializing grafics

	double alfa=0,beta=0,gamma=0;
		// angles used for rotating object
	double dalfa=0,dbeta=0,dgamma=0;	// how much to add every frame

	vertex lamp;			// defines direction of light source
		lamp.globalx=0;
		lamp.globaly=0;   	// this is position of a little cross (see
		lamp.globalz=-300;	// on the screen :) )
		lamp.normx = 0;
		lamp.normy = 0;
		lamp.normz = 1;

	double lamp_alfa=0,lamp_beta=0;
	double lamp_dalfa=M_PI/30,lamp_dbeta=M_PI/30;

	long a;	// guess what it's for :)

	for(a=0; a<no_of_faces; a++)
		calculate_N_vektor(&faces[a]);
		// let's add normal vector of each face to all vertices of the face

	for(a=0; a<no_of_vertices;a++)
		calculate_points_N_vektor(&vertices[a]);
		// and then let's make them be unit

	compute_env_map();

	clock_t cstart, cend;	// we use clock to measure the speed
	cstart = clock();       // let it start

	double frames = 0;		// no of frames we've generated yet
	while(1)			// till U hit esc
	{
		alfa  += dalfa;
		beta  += dbeta;		// modifies angles of object
		gamma += dgamma;

		rotate_lamp(&lamp,lamp_alfa,lamp_beta);
			// only around two axis

		rotate_vertices(vertices,no_of_vertices,alfa,beta,gamma);

		/* let's compute H vector. H = (L+V)/2 , V = [0,0,1] */
		Hx = lamp.nx/2;
		Hy = lamp.ny/2;
		Hz = (lamp.nz+1)/2;

		double len = sqrt(Hx*Hx+Hy*Hy+Hz*Hz);
		if(len)
		{
			Hx/=len;
			Hy/=len;
			Hz/=len;
		}
		rotate_Vnorm(vertices,no_of_vertices);

		for(a=0; a<no_of_vertices; a++)
			to_2d(&vertices[a]);

		long ile_sort = test_visibility(faces,faces_for_drawing,no_of_faces);

		qsort(faces_for_drawing,ile_sort,sizeof(sort_face),sortfn);
			// sort using sortfn



		if(lamp.z>0) draw_light(lamp);
		draw_faces(faces_for_drawing);
		if(lamp.z<=0) draw_light(lamp);

		while ( (inportb(0x03da) & 8) == 0);  //czekanie na ramke
		while ( (inportb(0x03da) & 8) == 8);
		to_screen();
		frames+=1;

		if(kbhit())
		{
			int c = getch();
			if( c=='a' )break;
			if( c=='k' )lamp_alfa += lamp_dalfa;
			if( c==';' )lamp_alfa -= lamp_dalfa;
			if( c=='o' )lamp_beta -= lamp_dbeta;
			if( c=='l' )lamp_beta += lamp_dbeta;
			if( c=='f' )dalfa += M_PI/100;
			if( c=='r' )dalfa -= M_PI/100;
			if( c=='g' )dbeta += M_PI/100;
			if( c=='t' )dbeta -= M_PI/100;
			if( c=='h' )dgamma += M_PI/100;
			if( c=='y' )dgamma -= M_PI/100;
			if( c=='a' )break;
		}
	}
	cend = clock();

	gfx_exit();


	printf("Average frames per second: %f\n", frames/((cend - cstart) / CLK_TCK));


	farfree(screen1);
	farfree(screen2);
	farfree(screen3);
	farfree(env_map);
	farfree(vertices);
	farfree(faces);
	farfree(faces_for_drawing);

//	cout << ile_punktow1 << "     " << ile_scian1;
//	cout << '\n' << max_punkty << "     " << max_sciany;
	getch();

	return 0;
}
matrix transform;

void rotate_vertices(vertex* vertices, long no,
			double alfa,double beta,double gamma)
{
	matrix Mz(cos(alfa), -sin(alfa), 0, 0,
				 sin(alfa),  cos(alfa), 0, 0,
				 0			 ,  0			, 1, 0,
				 0			 ,	 0			, 0, 1);

	matrix Mx(1, 0 		 , 0, 		0,
				 0, cos(beta),-sin(beta), 0,
				 0, sin(beta), cos(beta), 0,
				 0, 0			 , 0, 		1);

	matrix My(cos(gamma), 0, sin(gamma), 0,
				 0, 1, 0, 0,
				 -sin(gamma),0, cos(gamma), 0,
				 0, 0			 , 0, 		1);

	transform = My*Mx*Mz;

	for(int a=0; a<no; a++)
		transform.rotate(&vertices[a]);
	return;
}

void rotate_Vnorm(vertex* vertices, long no)
{

	vertex H;
		H.normx = Hx;
		H.normy = Hy;
		H.normz = Hz;
	transform.trans();		// reverse transformation


	transform.rotatenorm(&H);	// for H vector
	Hx = H.nx;
	Hy = H.ny;
	Hz = H.nz;


	double A=sqrt(H.nx*H.nx+H.nz*H.nz);
	if(A==0)A=1;

	H.normx = Hx;
	H.normy = Hy;
	H.normz = Hz;


	matrix M(H.nz/A,		0,    -H.nx/A,	     0,
				-H.nx*H.ny/A, A,    -H.nz*H.ny/A, 0,
				 H.nx,		H.ny, H.nz, 		  0,
				 0,		   0,    0,	    		  1);
	M.rotatenorm(&H);

	for(int a=0; a<no; a++)
		M.rotatenorm(&vertices[a]);
	return;
}




inline unsigned char phong(double Nx,double Ny)
{
	int x, y;
	x = Nx;
	y = Ny;

	 return env_map[256*x+y];
}

void draw_faces(sort_face faces[])
{
	for(int i=0;faces[i].address;i++)
		triangle(faces[i].address->P1,faces[i].address->P2,
												faces[i].address->P3);
	return;
}



double dnx,dny,dnz;		// these are the values we add every pixel in line
								// they are constant for entire triangle

void inline line(double xl,double nxl,double nyl,
			 double xp,double nxp,double nyp,
			 long y)
{
	unsigned char col;
	long x1 = xl;
	long x2 = xp;

	while(x1<x2)
	{

			// draws single pixel
			if (x1>0 && x1<320 && y>0 && y<200)
				screen2[y*320+x1] = phong(nxl,nyl);
		nxl+=dnx;
		nyl+=dny;
		x1++;
	}
}

void triangle(vertex* P1,vertex* P2,vertex* P3)
{

	// let's sort the vertices so that P1->y <= P2->y <= P3->y

	if(P1->drawy > P2->drawy)
	{	// change
		vertex* temp = P1;
		P1 = P2;
		P2 = temp;
	}
	if(P1->drawy > P3->drawy)
	{	// change
		vertex* temp = P1;
		P1 = P3;
		P3 = temp;
	}
	if(P2->drawy > P3->drawy)
	{	// change
		vertex* temp = P2;
		P2 = P3;
		P3 = temp;
	}
	// the smallest y is P1, then P2 and P3

	double P1nx = P1->nx*127.5+127.5;
	double P1ny = P1->ny*127.5+127.5;



	double P2nx = P2->nx*127.5+127.5;
	double P2ny = P2->ny*127.5+127.5;

	double P3nx = P3->nx*127.5+127.5;
	double P3ny = P3->ny*127.5+127.5;

	double dx12,dnx12,dny12;
	double dx13,dnx13,dny13;
	double dx23,dnx23,dny23;

	long dy12,dy13,dy23;

	dy12 = P2->drawy-P1->drawy;
	dy13 = P3->drawy-P1->drawy;
	dy23 = P3->drawy-P2->drawy;
	if(dy12>0)
	{
		dx12 = ((double)(P2->drawx-P1->drawx))/dy12;
		dnx12 = (P2nx-P1nx)/dy12;
		dny12 = (P2ny-P1ny)/dy12;
	}
	if(dy13>0)
	{
		dx13 = ((double)(P3->drawx-P1->drawx))/dy13;
		dnx13 = (P3nx-P1nx)/dy13;
		dny13 = (P3ny-P1ny)/dy13;
	}
	if(dy23>0)
	{
		dx23 = ((double)(P3->drawx-P2->drawx))/dy23;
		dnx23 = (P3nx-P2nx)/dy23;
		dny23 = (P3ny-P2ny)/dy23;
	}

	/* the addres for a scanline */
	/* first, let's compute P4
							P1
							/\
						  /  \
						 /	   \
				  P2	/		 \ <- P4
						 \      \
							\     \
							  \    \
								  \  \
									  \  P3									*/
	double p4x,p4nx,p4ny,p4nz;
	p4x = P1->drawx + dx13*dy12;
	p4nx = P1nx + dnx13*dy12;
	p4ny = P1ny + dny13*dy12;

	/* now the adders */
	double dx = p4x-P2->drawx;
	if(dx!=0)
	{
		dnx = (p4nx-P2nx)/dx;
		dny = (p4ny-P2ny)/dx;
	}
	else return; // don't need to draw anything

	double xl,nxl,nyl;
	double xp,nxp,nyp;

	double dxl,dnxl,dnyl;
	double dxp,dnxp,dnyp;

	xl = P1->drawx;
	nxl = P1nx;
	nyl = P1ny;

	xp = P1->drawx;
	nxp = P1nx;
	nyp = P1ny;

	dxl = dx12;
	dnxl = dnx12;
	dnyl = dny12;

	dxp = dx13;
	dnxp = dnx13;
	dnyp = dny13;

	long yline = P1->drawy;

	while(yline<P2->drawy)
	{
		if(xl<xp) line(xl,nxl,nyl,xp,nxp,nyp,yline);
		else line(xp,nxp,nyp,xl,nxl,nyl,yline);

		xl += dxl;
		nxl += dnxl;
		nyl += dnyl;

		xp += dxp;
		nxp += dnxp;
		nyp += dnyp;

		yline++;
	}

	xl = P2->drawx;		// this four lines are necessary when P1->y == P2->y
	nxl = P2nx;			// try to remove them if you don't know why :)
	nyl = P2ny;

	dxl = dx23;
	dnxl = dnx23;
	dnyl = dny23;

	while(yline<P3->drawy)
	{
		if(xl<xp) line(xl,nxl,nyl,xp,nxp,nyp,yline);
		else line(xp,nxp,nyp,xl,nxl,nyl,yline);

		xl += dxl;
		nxl += dnxl;
		nyl += dnyl;

		xp += dxp;
		nxp += dnxp;
		nyp += dnyp;

		yline++;
	}
	return;
}

