#include "datatype.h"

unsigned char far* env_map;
unsigned char far* screen1;
unsigned char far* screen2;
unsigned char far* screen3;

#include "common.h"

#include "initcode.h"

inline void rotate_vertices(vertex*,long,double,double,double);
void draw_faces(sort_face*,vertex*);
void triangle(vertex*,vertex*,vertex*,vertex*);

int main(void)
{
	/* Let's allocate some memory for our structures and data */
	screen1 = (unsigned char far*) farcalloc(64000l,1);
	screen2 = (unsigned char far*) farcalloc(64500l,1);
	screen3 = (unsigned char far*) farcalloc(64500l,1);
	if(!screen1) error("not enought memory");
	if(!screen2) error("not enought memory");
	if(!screen3) error("not enought memory");


	const long max_vertices = 65000l/sizeof(vertex);
		// maximum amount of vertices that can be stored in memory
	const long max_faces = 65000l/sizeof(face);
		// maximum amount of faces that can be stored in memory


	vertex far* vertices = (vertex far*)farcalloc(max_vertices,sizeof(vertex) );
		/* mem for all the vertices */

	face far* faces = (face far*)farcalloc(max_faces,sizeof(face) );
		/* and faces */

	sort_face far* faces_for_drawing = (sort_face far*)
								farcalloc(max_faces+1,sizeof(sort_face) );
		/* this is necessary when sorting faces */
	if(!vertices)  error("not enought memory");
	if(!faces)  error("not enought memory");
	if(!faces_for_drawing)  error("not enought memory");

	long no_of_vertices=0,no_of_faces=0;

   read_config();
	if(objraw)
		read_raw(vertices,max_vertices,faces,max_faces,
					no_of_vertices,no_of_faces);
	/* function modifies the variables no_of_faces and no_of_vertices
		while reading data from file */
	else
		read_file(vertices,max_vertices,faces,max_faces,
					no_of_vertices,no_of_faces);

	read_picture(screen3);
		/* reads picture for background to screen3 pointer */

	if ( !no_of_vertices || !no_of_faces ) error("plik uszkodzony");
		// if no vertices of no faces read
	init_gfx();

	setup_palette();	// must be after initializing grafics

	double alfa=0,beta=0,gamma=0;
		// angles used for rotating object
	double dalfa=0,dbeta=0,dgamma=0;	// how much to add every frame

	vertex lamp;			// defines direction of light source
		lamp.globalx=0;
		lamp.globaly=0;   	// this is position of a little cross (see
		lamp.globalz=-300;	// on the screen :) )
		lamp.normx = 0;
		lamp.normy = 0;
		lamp.normz = 1;

	double lamp_alfa=0,lamp_beta=0;
	double lamp_dalfa=M_PI/30,lamp_dbeta=M_PI/30;

	long a;	// guess what it's for :)

	for(a=0; a<no_of_faces; a++)
		calculate_N_vektor(&faces[a]);
		// let's add normal vector of each face to all vertices of the face

	for(a=0; a<no_of_vertices;a++)
		calculate_points_N_vektor(&vertices[a]);
		// and then let's make them be unit

	clock_t cstart, cend;	// we use clock to measure the speed
	cstart = clock();       // let it start

	double frames = 0;		// no of frames we've generated yet
	while(1)			// till U hit esc
	{
		alfa  += dalfa;
		beta  += dbeta;		// modifies angles of object
		gamma += dgamma;

		rotate_lamp(&lamp,lamp_alfa,lamp_beta);
			// only around two axis

		rotate_vertices(vertices,no_of_vertices,alfa,beta,gamma);

		for(a=0; a<no_of_vertices; a++)
			to_2d(&vertices[a]);

		long ile_sort = test_visibility(faces,faces_for_drawing,no_of_faces);


		qsort(faces_for_drawing,ile_sort,sizeof(sort_face),sortfn);
			// sort using sortfn

		if(lamp.z>0) draw_light(lamp);
		draw_faces(faces_for_drawing,&lamp);
		if(lamp.z<=0) draw_light(lamp);

		while ( (inportb(0x03da) & 8) == 0);  //czekanie na ramke
		while ( (inportb(0x03da) & 8) == 8);
		to_screen();
		frames+=1;

		if(kbhit())
		{
			int c = getch();
			if( c=='a' )break;
			if( c=='k' )lamp_alfa += lamp_dalfa;
			if( c==';' )lamp_alfa -= lamp_dalfa;
			if( c=='o' )lamp_beta -= lamp_dbeta;
			if( c=='l' )lamp_beta += lamp_dbeta;
			if( c=='f' )dalfa += M_PI/100;
			if( c=='r' )dalfa -= M_PI/100;
			if( c=='g' )dbeta += M_PI/100;
			if( c=='t' )dbeta -= M_PI/100;
			if( c=='h' )dgamma += M_PI/100;
			if( c=='y' )dgamma -= M_PI/100;
			if( c=='a' )break;
		}
	}
	cend = clock();

	gfx_exit();

	printf("Average frames per second: %f\n", frames/((cend - cstart) / CLK_TCK));

	farfree(screen1);
	farfree(screen2);
	farfree(screen3);
	farfree(env_map);
	farfree(vertices);
	farfree(faces);
	farfree(faces_for_drawing);
	getch();

	return 0;
}

void rotate_vertices(vertex* vertices, long no,
			double alfa,double beta,double gamma)
{
	matrix Mz(cos(alfa), -sin(alfa), 0, 0,
				 sin(alfa),  cos(alfa), 0, 0,
				 0			 ,  0			, 1, 0,
				 0			 ,	 0			, 0, 1);

	matrix Mx(1, 0 		 , 0, 		0,
				 0, cos(beta),-sin(beta), 0,
				 0, sin(beta), cos(beta), 0,
				 0, 0			 , 0, 		1);

	matrix My(cos(gamma), 0, sin(gamma), 0,
				 0, 1, 0, 0,
				 -sin(gamma),0, cos(gamma), 0,
				 0, 0			 , 0, 		1);

	matrix transform = My*Mx*Mz;

		for(int a=0; a<no; a++)
		{
			transform.rotate(&vertices[a]);
			transform.rotatenorm(&vertices[a]);
		}
		return;
}

unsigned char phong(double Nx,double Ny,double Nz,vertex* L)
{
	// original phong shading method
	double d = sqrt(Nx*Nx+Ny*Ny+Nz*Nz);
	if (d)
	{
		Nx/=d;
		Ny/=d;
		Nz/=d;
	}

	static const double m = 12;
	// Phong
	// observer V = {0,0,1}

	//n o l
	double nol = Nx*L->nx + Ny*L->ny + Nz*L->nz;
	// Rl o V - refracted light and surface
	// Rl = 2N(L o N)-L
	// Rl o V = V o( 2N(L o N)-L ) = 2(L o N)(V o N) - V o L
	// V o N = 1 * Nz
	// V o L = 1 * Lz
	double vorl;

	vorl = 2*nol*Nz - L->nz;

	int col=0;

	if (vorl>1) vorl = 1;

	double vorl1;

	if (nol>0)	// if the light is not behind the surface
	{				// it can affect the colour of the pixel
		col+=nol*40;
		if (vorl>0)
		{
			vorl1 = pow(vorl,m);
			col += vorl1*250;
		}
	}
	if(col<250)
		return col;
	else
		return 250;
}


void draw_faces(sort_face faces[],vertex* L)
{
	for(int i=0;faces[i].address;i++)
		triangle(faces[i].address->P1,faces[i].address->P2,
												faces[i].address->P3,L);
	return;
}



double dnx,dny,dnz;		// these are the values we add every pixel in line
								// they are constant for entire triangle

void inline line(double xl,double nxl,double nyl,double nzl,
			 double xp,double nxp,double nyp,double nzp,
			 long y,vertex* L)
{
	unsigned char col;
	long x1 = xl;
	long x2 = xp;

	while(x1<x2)
	{

			// draws single pixel
			if (x1>0 && x1<320 && y>0 && y<200)
				screen2[y*320+x1] = phong(nxl,nyl,nzl,L);
		nxl+=dnx;
		nyl+=dny;
		nzl+=dnz;
		x1++;
	}
}

void triangle(vertex* P1,vertex* P2,vertex* P3,vertex* L)
{

	// let's sort the vertices so that P1->y <= P2->y <= P3->y

	if(P1->drawy > P2->drawy)
	{	// change
		vertex* temp = P1;
		P1 = P2;
		P2 = temp;
	}
	if(P1->drawy > P3->drawy)
	{	// change
		vertex* temp = P1;
		P1 = P3;
		P3 = temp;
	}
	if(P2->drawy > P3->drawy)
	{	// change
		vertex* temp = P2;
		P2 = P3;
		P3 = temp;
	}
	// the smallest y is P1, then P2 and P3

	double dx12,dnx12,dny12,dnz12;
	double dx13,dnx13,dny13,dnz13;
	double dx23,dnx23,dny23,dnz23;

	long dy12,dy13,dy23;

	dy12 = P2->drawy-P1->drawy;
	dy13 = P3->drawy-P1->drawy;
	dy23 = P3->drawy-P2->drawy;
	if(dy12>0)
	{
		dx12 = ((double)(P2->drawx-P1->drawx))/dy12;
		dnx12 = (P2->nx-P1->nx)/dy12;
		dny12 = (P2->ny-P1->ny)/dy12;
		dnz12 = (P2->nz-P1->nz)/dy12;
	}
	if(dy13>0)
	{
		dx13 = ((double)(P3->drawx-P1->drawx))/dy13;
		dnx13 = (P3->nx-P1->nx)/dy13;
		dny13 = (P3->ny-P1->ny)/dy13;
		dnz13 = (P3->nz-P1->nz)/dy13;
	}
	if(dy23>0)
	{
		dx23 = ((double)(P3->drawx-P2->drawx))/dy23;
		dnx23 = (P3->nx-P2->nx)/dy23;
		dny23 = (P3->ny-P2->ny)/dy23;
		dnz23 = (P3->nz-P2->nz)/dy23;
	}

	/* the addres for a scanline */
	/* first, let's compute P4
							P1
							/\
						  /  \
						 /	   \
				  P2	/		 \ <- P4
						 \      \
							\     \
							  \    \
								  \  \
									  \  P3									*/
	double p4x,p4nx,p4ny,p4nz;
	p4x = P1->drawx + dx13*dy12;
	p4nx = P1->nx + dnx13*dy12;
	p4ny = P1->ny + dny13*dy12;
	p4nz = P1->nz + dnz13*dy12;

	/* now the adders */
	double dx = p4x-P2->drawx;
	if(dx!=0)
	{
		dnx = (p4nx-P2->nx)/dx;
		dny = (p4ny-P2->ny)/dx;
		dnz = (p4nz-P2->nz)/dx;
	}
	else return; // don't need to draw anything

	double xl,nxl,nyl,nzl;
	double xp,nxp,nyp,nzp;

	double dxl,dnxl,dnyl,dnzl;
	double dxp,dnxp,dnyp,dnzp;

	xl = P1->drawx;
	nxl = P1->nx;
	nyl = P1->ny;
	nzl = P1->nz;

	xp = P1->drawx;
	nxp = P1->nx;
	nyp = P1->ny;
	nzp = P1->nz;

	dxl = dx12;
	dnxl = dnx12;
	dnyl = dny12;
	dnzl = dnz12;

	dxp = dx13;
	dnxp = dnx13;
	dnyp = dny13;
	dnzp = dnz13;

	long yline = P1->drawy;

	while(yline<P2->drawy)
	{
		if(xl<xp) line(xl,nxl,nyl,nzl,xp,nxp,nyp,nzp,yline,L);
		else line(xp,nxp,nyp,nzp,xl,nxl,nyl,nzl,yline,L);

		xl += dxl;
		nxl += dnxl;
		nyl += dnyl;
		nzl += dnzl;

		xp += dxp;
		nxp += dnxp;
		nyp += dnyp;
		nzp += dnzp;

		yline++;
	}

	xl = P2->drawx;		// this four lines are necessary when P1->y == P2->y
	nxl = P2->nx;			// try to remove them if you don't know why :)
	nyl = P2->ny;
	nzl = P2->nz;

	dxl = dx23;
	dnxl = dnx23;
	dnyl = dny23;
	dnzl = dnz23;

	while(yline<P3->drawy)
	{
		if(xl<xp) line(xl,nxl,nyl,nzl,xp,nxp,nyp,nzp,yline,L);
		else line(xp,nxp,nyp,nzp,xl,nxl,nyl,nzl,yline,L);

		xl += dxl;
		nxl += dnxl;
		nyl += dnyl;
		nzl += dnzl;

		xp += dxp;
		nxp += dnxp;
		nyp += dnyp;
		nzp += dnzp;

		yline++;
	}
	return;
}




